from rollerverbrauch import db
from flask.ext.security import UserMixin, RoleMixin

roles_users = db.Table('roles_users',
                       db.Column('user_id', db.Integer(), db.ForeignKey('user.id')),
                       db.Column('role_id', db.Integer(), db.ForeignKey('role.id')))

vehicles_consumables = db.Table('vehicles_consumables',
                                db.Column('vehicle_id', db.Integer(), db.ForeignKey('vehicle.id')),
                                db.Column('consumable_id', db.Integer(), db.ForeignKey('consumable.id')))


class Role(db.Model, RoleMixin):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(80), unique=True)
    description = db.Column(db.String(255))

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)


class User(db.Model, UserMixin):
    id = db.Column(db.Integer, primary_key=True)
    email = db.Column(db.String(255), unique=True)
    password = db.Column(db.String(255))
    active = db.Column(db.Boolean())
    confirmed_at = db.Column(db.DateTime())
    vehicles = db.relationship(
        'Vehicle'
    )
    roles = db.relationship(
        'Role',
        secondary=roles_users,
        backref=db.backref('users', lazy='dynamic')
    )

    def __repr__(self):
        return '<User id="%r" email="%r" ' % (self.id, self.email)


class Vehicle(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    owner_id = db.Column(db.Integer, db.ForeignKey('user.id'))
    name = db.Column(db.String(255))
    pitstops = db.relationship(
        'Pitstop'
    )
    consumables = db.relationship(
        'Consumable',
        secondary=vehicles_consumables,
        backref=db.backref('consumes', lazy='dynamic')
    )
    __table_args__ = (db.UniqueConstraint('owner_id', 'name', name='_owner_name_uniq'),)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<Vehicle id="%r" owner_id="%r" name="%r" />' % (self.id, self.owner_id, self.name)


class Pitstop(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    date = db.Column(db.Date)
    odometer = db.Column(db.Integer)
    consumable_id = db.Column(db.Integer, db.ForeignKey('consumable.id'))
    consumable = db.relationship('Consumable')
    litres = db.Column(db.Numeric(5, 2))
    costs = db.Column(db.Numeric(5, 2), default=0)
    vehicle_id = db.Column(db.Integer, db.ForeignKey('vehicle.id'))

    def __init__(self, odometer, litres, date, costs, consumable_id):
        self.odometer = odometer
        self.litres = litres
        self.date = date
        self.costs = costs
        self.consumable_id = consumable_id

    def __repr__(self):
        return '<Pitstop odometer="%r" litres="%r" date="%r" vehicle_id="%r" consumable_id="%r">' % \
               (self.odometer, self.litres, self.date, self.vehicle_id, self.consumable_id)


class Consumable(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), unique=True)
    unit = db.Column(db.String(255))

    def __init__(self, name, unit):
        self.name = name
        self.unit = unit

    def __repr__(self):
        return '<Consumable name="%s" unit="%s" />' % (self.name, self.unit)
